#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/html/win32/hos_w32.cpp,v 1.3 1999/05/29 15:51:04 MJRoberts Exp $";
#endif

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  hos_w32.cpp - HTML OS functions for Win32
Function
  
Notes
  
Modified
  01/24/98 MJRoberts  - Creation
*/

#include <Windows.h>

#include <os.h>

#ifndef HTML_OS_H
#include "html_os.h"
#endif
#ifndef TADSHTML_H
#include "tadshtml.h"
#endif
#ifndef HTMLW32_H
#include "htmlw32.h"
#endif
#ifndef HTMLPREF_H
#include "htmlpref.h"
#endif
#ifndef W32MAIN_H
#include "w32main.h"
#endif


/* ------------------------------------------------------------------------ */
/*
 *   get the current system time
 */
os_timer_t os_get_time()
{
    return GetTickCount();
}


/* ------------------------------------------------------------------------ */
/*
 *   Get system information 
 */
int os_get_sysinfo(int code, void * /*param*/, long *result)
{
    CHtmlSys_mainwin *win;
    CHtmlPreferences *prefs;
    
    /* get the preferences object, in case we need it */
    if ((win = CHtmlSys_mainwin::get_main_win()) != 0)
        prefs = win->get_prefs();
    else
        prefs = 0;

    /* find out what kind of information they want */
    switch(code)
    {
    case SYSINFO_HTML:
    case SYSINFO_JPEG:
    case SYSINFO_PNG:
    case SYSINFO_MIDI:
        /* 
         *   we definitely support HTML, JPEG, PNG, and MIDI 
         */
        *result = 1;
        return TRUE;

    case SYSINFO_WAV:
    case SYSINFO_WAV_MIDI_OVL:
    case SYSINFO_WAV_OVL:
    case SYSINFO_MPEG:
    case SYSINFO_MPEG2:
    case SYSINFO_MPEG3:
        /*
         *   We may support WAV or not, depending on DirectX availability.
         *   If WAV isn't supported, then the overlap features aren't
         *   meaningful, but return false for them anyway in that case.
         *   We support WAV and the various overlapped playback features
         *   if and only if DirectSound is available.
         *   
         *   We use an integrated software player to play MPEG layer
         *   II and III sounds.  However, we essentially decode MPEG
         *   sounds to WAV in memory and play the resulting WAV through
         *   DirectX, so we require DirectX to play MPEG's, just as we do
         *   for WAV's.  
         */
        *result = (win->get_directsound() != 0);
        return TRUE;

    case SYSINFO_PREF_IMAGES:
        *result = (prefs != 0 && prefs->get_graphics_on() ? 1 : 0);
        return TRUE;
        
    case SYSINFO_PREF_SOUNDS:
        *result = (prefs != 0 && prefs->get_sounds_on() ? 1 : 0);
        return TRUE;
        
    case SYSINFO_PREF_MUSIC:
        *result = (prefs != 0 && prefs->get_music_on() ? 1 : 0);
        return TRUE;
        
    case SYSINFO_PREF_LINKS:
        *result = (prefs != 0
                   ? (prefs->get_links_on()
                      ? 1
                      : prefs->get_links_ctrl() ? 2 : 0)
                   : 0);
        return TRUE;

    case SYSINFO_LINKS_HTTP:
    case SYSINFO_LINKS_FTP:
    case SYSINFO_LINKS_NEWS:
    case SYSINFO_LINKS_MAILTO:
    case SYSINFO_LINKS_TELNET:
        *result = TRUE;
        return TRUE;

    default:
        /* we don't recognize other codes */
        return FALSE;
    }
}


/* ------------------------------------------------------------------------ */
/*
 *   Receive notification that a character mapping file has been loaded.
 */
void os_advise_load_charmap(char *id, char *ldesc, char *sysinfo)
{
    CHtmlSys_mainwin *mainwin;
    
    /* get the main window, and tell it about the change */
    mainwin = (CHtmlSys_mainwin *)oshtml_get_frame();
    if (mainwin != 0)
        mainwin->advise_load_charmap(ldesc, sysinfo);
}


/* ------------------------------------------------------------------------ */
/*
 *   Terminate the program 
 */
void os_term(int exit_code)
{
    /* post a quit message */
    PostMessage(0, WM_QUIT, exit_code, 0);
}

